@@ -90,10 +90,17 @@ def infer_json_schema(
     schema_uri: str | None = "http://json-schema.org/schema#",
     merge_schemas: bool = True,
     debug: bool = False,
+    profile: bool = False,
+    verbosity: Literal["Normal", "Verbose"] = "Normal",
     map_threshold: int = 20,
+    map_max_required_keys: int | None = None,
+    unify_maps: bool = False,
+    no_unify: set[str] | None = None,
     force_field_types: dict[str, str] | None = None,
+    wrap_scalars: bool = True,
     avro: bool = False,
     wrap_root: str | None = None,
+    no_root_map: bool = True,
 ) -> pl.Expr:
     """Infer JSON schema from a string column containing JSON data.
 
@@ -111,17 +118,39 @@ def infer_json_schema(
         Whether to merge schemas from all rows (True) or return individual schemas (False)
     debug : bool, default False
         Whether to print debug information
+    profile : bool, default False
+        Whether to print profiling information
+    verbosity : str, default "Normal"
+        Whether to print verbose debug information
     map_threshold : int, default 20
         Number of keys above which a heterogeneous object may be rewritten
         as a map (unless overridden).
+    map_max_required_keys : int, optional
+        Maximum number of required keys allowed for Map inference. Objects with more
+        required keys will be forced to Record type. If None, no gating based on
+        required key count (preserves existing behavior).
+    unify_maps : bool, default False
+        Enable unification of compatible but non-homogeneous record schemas into maps.
+        When True, record schemas with compatible field types can be merged into a single
+        map schema with selective nullable fields.
+    no_unify: set[str] | None, default None
+        Prevent unification of keys under these field names with their sibling record fields.
     force_field_types : dict[str, str], optional
         Explicit overrides for specific fields. Values must be `"map"` or `"record"`.
         Example: ``{"labels": "map", "claims": "record"}``.
+    wrap_scalars : bool, default True
+        Whether to promote scalar values into singleton objects when they appear
+        in contexts where other rows provide objects. This avoids unification
+        failures between scalars and objects. The promoted field name defaults
+        to the parent key with a ``__{type}`` suffix, e.g. a string under
+        ``"value"`` becomes ``{"value__string": "..."}``.
     avro: bool, default False
         Whether to output an Avro schema instead of JSON schema.
     wrap_root : str | None, default None
         If a string, wrap each JSON row under that key before inference.
         If ``None``, leave rows unchanged.
+    no_root_map : bool, default True
+        Prevent document root from becoming a map type, even if it meets map inference criteria
 
     Returns:
     -------
@@ -133,9 +162,16 @@ def infer_json_schema(
         "ndjson": ndjson,
         "merge_schemas": merge_schemas,
         "debug": debug,
+        "profile": profile,
+        "verbosity": verbosity,
         "map_threshold": map_threshold,
+        "map_max_required_keys": map_max_required_keys,
+        "unify_maps": unify_maps,
+        "no_unify": list(no_unify) if no_unify else [],
+        "wrap_scalars": wrap_scalars,
         "avro": avro,
         "wrap_root": wrap_root,
+        "no_root_map": no_root_map,
     }
     if schema_uri is not None:
         kwargs["schema_uri"] = schema_uri
@@ -152,10 +188,17 @@ def infer_polars_schema(
     ndjson: bool = False,
     merge_schemas: bool = True,
     debug: bool = False,
+    profile: bool = False,
+    verbosity: Literal["Normal", "Verbose"] = "Normal",
     map_threshold: int = 20,
+    map_max_required_keys: int | None = None,
+    unify_maps: bool = False,
+    no_unify: set[str] | None = None,
     force_field_types: dict[str, str] | None = None,
+    wrap_scalars: bool = True,
     avro: bool = False,
     wrap_root: str | None = None,
+    no_root_map: bool = True,
 ) -> pl.Expr:
     """Infer Polars schema from a string column containing JSON data.
 
@@ -171,17 +214,39 @@ def infer_polars_schema(
         Whether to merge schemas from all rows (True) or return individual schemas (False)
     debug : bool, default False
         Whether to print debug information
+    profile : bool, default False
+        Whether to print profiling information
+    verbosity : str, default "Normal"
+        Whether to print verbose debug information
     map_threshold : int, default 20
         Number of keys above which a heterogeneous object may be rewritten
         as a map (unless overridden).
+    map_max_required_keys : int, optional
+        Maximum number of required keys allowed for Map inference. Objects with more
+        required keys will be forced to Record type. If None, no gating based on
+        required key count.
+    unify_maps : bool, default False
+        Enable unification of compatible but non-homogeneous record schemas into maps.
+        When True, record schemas with compatible field types can be merged into a single
+        map schema with selective nullable fields.
+    no_unify: set[str] | None, default None
+        Prevent unification of keys under these field names with their sibling record fields.
     force_field_types : dict[str, str], optional
         Explicit overrides for specific fields. Values must be `"map"` or `"record"`.
         Example: ``{"labels": "map", "claims": "record"}``.
+    wrap_scalars : bool, default True
+        Whether to promote scalar values into singleton objects when they appear
+        in contexts where other rows provide objects. This avoids unification
+        failures between scalars and objects. The promoted field name defaults
+        to the parent key with a ``__{type}`` suffix, e.g. a string under
+        ``"value"`` becomes ``{"value__string": "..."}``.
     avro: bool, default False
         Whether to read the input as an Avro schema instead of JSON schema.
     wrap_root : str | None, default None
         If a string, wrap each JSON row under that key before inference.
         If ``None``, leave rows unchanged.
+    no_root_map : bool, default True
+        Prevent document root from becoming a map type, even if it meets map inference criteria
 
     Returns:
     -------
@@ -193,9 +258,16 @@ def infer_polars_schema(
         "ndjson": ndjson,
         "merge_schemas": merge_schemas,
         "debug": debug,
+        "profile": profile,
+        "verbosity": verbosity,
         "map_threshold": map_threshold,
+        "map_max_required_keys": map_max_required_keys,
+        "unify_maps": unify_maps,
+        "no_unify": list(no_unify) if no_unify else [],
+        "wrap_scalars": wrap_scalars,
         "avro": avro,
         "wrap_root": wrap_root,
+        "no_root_map": no_root_map,
     }
     if not merge_schemas:
         url = "https://github.com/lmmx/polars-genson/issues/37"
@@ -214,9 +286,15 @@ def normalise_json(
     empty_as_null: bool = True,
     coerce_strings: bool = False,
     map_encoding: Literal["entries", "mapping", "kv"] = "kv",
+    profile: bool = False,
     map_threshold: int = 20,
+    map_max_required_keys: int | None = None,
+    unify_maps: bool = False,
+    no_unify: set[str] | None = None,
     force_field_types: dict[str, str] | None = None,
+    wrap_scalars: bool = True,
     wrap_root: str | None = None,
+    no_root_map: bool = True,
 ) -> pl.Expr:
     """Normalise a JSON string column against an inferred Avro schema.
 
@@ -243,15 +321,35 @@ def normalise_json(
         - "mapping": plain JSON object ({"en":"Hello"})
         - "entries": list of single-entry objects ([{"en":"Hello"}])
         - "kv":      list of {key,value} dicts ([{"key":"en","value":"Hello"}])
+    profile : bool, default False
+        Whether to show timing profile output
     map_threshold : int, default 20
         Maximum number of keys before an object is treated as a map
         (unless overridden).
+    map_max_required_keys : int, optional
+        Maximum number of required keys allowed for Map inference during schema
+        inference. Objects with more required keys will be forced to Record type.
+        If None, no gating based on required key count.
+    unify_maps : bool, default False
+        Enable unification of compatible but non-homogeneous record schemas into maps.
+        When True, record schemas with compatible field types can be merged into a single
+        map schema with selective nullable fields.
+    no_unify: set[str] | None, default None
+        Prevent unification of keys under these field names with their sibling record fields.
     force_field_types : dict[str, str], optional
         Override the inferred type for specific fields. Keys are field names,
         values must be either ``"map"`` or ``"record"``.
+    wrap_scalars : bool, default True
+        Whether to promote scalar values into singleton objects when they appear
+        in contexts where other rows provide objects. This avoids unification
+        failures between scalars and objects. The promoted field name defaults
+        to the parent key with a ``__{type}`` suffix, e.g. a string under
+        ``"value"`` becomes ``{"value__string": "..."}``.
     wrap_root : str | None, default None
         Wrap each JSON row under that key before normalisation.
         If ``None``, leave rows unchanged.
+    no_root_map : bool, default True
+        Prevent document root from becoming a map type, even if it meets map inference criteria
 
     Returns:
     -------
@@ -284,8 +382,14 @@ def normalise_json(
         "empty_as_null": empty_as_null,
         "coerce_string": coerce_strings,
         "map_encoding": map_encoding,
+        "profile": profile,
         "map_threshold": map_threshold,
+        "map_max_required_keys": map_max_required_keys,
+        "unify_maps": unify_maps,
+        "no_unify": list(no_unify) if no_unify else [],
+        "wrap_scalars": wrap_scalars,
         "wrap_root": wrap_root,
+        "no_root_map": no_root_map,
     }
     if force_field_types is not None:
         kwargs["force_field_types"] = force_field_types
@@ -318,10 +422,17 @@ class GensonNamespace:
         ndjson: bool = False,
         merge_schemas: bool = True,
         debug: bool = False,
+        profile: bool = False,
+        verbosity: Literal["Normal", "Verbose"] = "Normal",
         map_threshold: int = 20,
+        map_max_required_keys: int | None = None,
+        unify_maps: bool = False,
+        no_unify: set[str] | None = None,
         force_field_types: dict[str, str] | None = None,
+        wrap_scalars: bool = True,
         avro: bool = False,
         wrap_root: bool | str | None = None,
+        no_root_map: bool = True,
     ) -> pl.Schema:
         # ) -> pl.Schema | list[pl.Schema]:
         """Infer Polars schema from a string column containing JSON data.
@@ -338,18 +449,40 @@ class GensonNamespace:
             Whether to merge schemas from all rows (True) or return individual schemas (False)
         debug : bool, default False
             Whether to print debug information
+        profile : bool, default False
+            Whether to print profiling information
+        verbosity : str, default "Normal"
+            Whether to print verbose debug information
         map_threshold : int, default 20
             Number of keys above which a heterogeneous object may be rewritten
             as a map (unless overridden).
+        map_max_required_keys : int, optional
+            Maximum number of required keys allowed for Map inference. Objects with more
+            required keys will be forced to Record type. If None, no gating based on
+            required key count.
+        unify_maps : bool, default False
+            Enable unification of compatible but non-homogeneous record schemas into maps.
+            When True, record schemas with compatible field types can be merged into a single
+            map schema with selective nullable fields.
+        no_unify: set[str] | None, default None
+            Prevent unification of keys under these field names with their sibling record fields.
         force_field_types : dict[str, str], optional
             Explicit overrides for specific fields. Values must be `"map"` or `"record"`.
             Example: ``{"labels": "map", "claims": "record"}``.
+        wrap_scalars : bool, default True
+            Whether to promote scalar values into singleton objects when they appear
+            in contexts where other rows provide objects. This avoids unification
+            failures between scalars and objects. The promoted field name defaults
+            to the parent key with a ``__{type}`` suffix, e.g. a string under
+            ``"value"`` becomes ``{"value__string": "..."}``.
         avro : bool, default False
             Whether to infer using Avro schema semantics (unions, maps, nullability).
             By default (`False`), JSON Schema mode is used.
         wrap_root : str | bool | None, default None
             If a string, wrap each JSON row under that key before inference.
             If ``True``, wrap under the column name. If ``None``, leave rows unchanged.
+        no_root_map : bool, default True
+            Prevent document root from becoming a map type, even if it meets map inference criteria
 
         Returns:
         -------
@@ -371,10 +504,16 @@ class GensonNamespace:
                 ndjson=ndjson,
                 merge_schemas=merge_schemas,
                 debug=debug,
+                profile=profile,
+                verbosity=verbosity,
                 map_threshold=map_threshold,
+                map_max_required_keys=map_max_required_keys,
+                unify_maps=unify_maps,
                 **fft,
+                wrap_scalars=wrap_scalars,
                 avro=avro,
                 wrap_root=wrap_root_field,
+                no_root_map=no_root_map,
             ).first()
         )
 
@@ -396,10 +535,17 @@ class GensonNamespace:
         schema_uri: str | None = "http://json-schema.org/schema#",
         merge_schemas: bool = True,
         debug: bool = False,
+        profile: bool = False,
+        verbosity: Literal["Normal", "Verbose"] = "Normal",
         map_threshold: int = 20,
+        map_max_required_keys: int | None = None,
+        unify_maps: bool = False,
+        no_unify: set[str] | None = None,
         force_field_types: dict[str, str] | None = None,
+        wrap_scalars: bool = True,
         avro: bool = False,
         wrap_root: bool | str | None = None,
+        no_root_map: bool = True,
     ) -> dict | list[dict]:
         """Infer JSON schema from a string column containing JSON data.
 
@@ -417,17 +563,39 @@ class GensonNamespace:
             Whether to merge schemas from all rows (True) or return individual schemas (False)
         debug : bool, default False
             Whether to print debug information
+        profile : bool, default False
+            Whether to print profiling information
+        verbosity : str, default "Normal"
+            Whether to print verbose debug information
         map_threshold : int, default 20
             Number of keys above which a heterogeneous object may be rewritten
             as a map (unless overridden).
+        map_max_required_keys : int, optional
+            Maximum number of required keys allowed for Map inference. Objects with more
+            required keys will be forced to Record type. If None, no gating based on
+            required key count.
+        unify_maps : bool, default False
+            Enable unification of compatible but non-homogeneous record schemas into maps.
+            When True, record schemas with compatible field types can be merged into a single
+            map schema with selective nullable fields.
+        no_unify: set[str] | None, default None
+            Prevent unification of keys under these field names with their sibling record fields.
         force_field_types : dict[str, str], optional
             Explicit overrides for specific fields. Values must be `"map"` or `"record"`.
             Example: ``{"labels": "map", "claims": "record"}``.
+        wrap_scalars : bool, default True
+            Whether to promote scalar values into singleton objects when they appear
+            in contexts where other rows provide objects. This avoids unification
+            failures between scalars and objects. The promoted field name defaults
+            to the parent key with a ``__{type}`` suffix, e.g. a string under
+            ``"value"`` becomes ``{"value__string": "..."}``.
         avro: bool, default False
             Whether to read the input as an Avro schema instead of JSON schema.
         wrap_root : str | bool | None, default None
             If a string, wrap each JSON row under that key before inference.
             If ``True``, wrap under the column name. If ``None``, leave rows unchanged.
+        no_root_map : bool, default True
+            Prevent document root from becoming a map type, even if it meets map inference criteria
 
         Returns:
         -------
@@ -444,10 +612,16 @@ class GensonNamespace:
                 schema_uri=schema_uri,
                 merge_schemas=merge_schemas,
                 debug=debug,
+                profile=profile,
+                verbosity=verbosity,
                 map_threshold=map_threshold,
+                map_max_required_keys=map_max_required_keys,
+                unify_maps=unify_maps,
                 force_field_types=force_field_types,
+                wrap_scalars=wrap_scalars,
                 avro=avro,
                 wrap_root=wrap_root_field,
+                no_root_map=no_root_map,
             ).first()
         )
 
@@ -465,16 +639,22 @@ class GensonNamespace:
         self,
         column: str,
         *,
-        decode: bool = True,
+        decode: bool | pl.Schema = True,
         unnest: bool = True,
         ignore_outer_array: bool = True,
         ndjson: bool = False,
         empty_as_null: bool = True,
         coerce_strings: bool = False,
         map_encoding: Literal["entries", "mapping", "kv"] = "kv",
+        profile: bool = False,
         map_threshold: int = 20,
+        map_max_required_keys: int | None = None,
+        unify_maps: bool = False,
+        no_unify: set[str] | None = None,
         force_field_types: dict[str, str] | None = None,
+        wrap_scalars: bool = True,
         wrap_root: bool | str | None = None,
+        no_root_map: bool = True,
     ) -> pl.Series:
         """Normalise a JSON string column to conform to an inferred Avro schema.
 
@@ -485,9 +665,15 @@ class GensonNamespace:
         ----------
         column : str
             Name of the column containing JSON strings.
-        decode : bool, default True
-            If True, decode the normalised JSON strings into native Polars
-            datatypes after normalisation. If False, leave as raw JSON strings instead.
+        decode : bool | pl.Schema, default True
+            Controls how the normalised JSON strings are decoded after
+            normalisation:
+
+            - If False leave values as raw JSON strings.
+            - If True (default), decode into native Polars datatypes,
+              with the schema inferred from the data (may be slower).
+            - If a polars.Schema, decode using the
+              provided schema dtype directly (fast path, skips final schema inference).
         unnest : bool, default True
             Only applies if `decode=True`. If True, expand the decoded struct
             into separate columns for each schema field. If False, keep a
@@ -508,14 +694,34 @@ class GensonNamespace:
             - "mapping": plain JSON object ({"en":"Hello"})
             - "entries": list of single-entry objects ([{"en":"Hello"}])
             - "kv":      list of {key,value} dicts ([{"key":"en","value":"Hello"}])
+        profile : bool, default False
+            Whether to display timing profile information
         map_threshold : int, default 20
             Threshold above which objects with many varying keys are normalised
             as Avro maps instead of records.
+        map_max_required_keys : int, optional
+            Maximum number of required keys allowed for Map inference during schema
+            inference. Objects with more required keys will be forced to Record type.
+            If None, no gating based on required key count.
+        unify_maps : bool, default False
+            Enable unification of compatible but non-homogeneous record schemas into maps.
+            When True, record schemas with compatible field types can be merged into a single
+            map schema with selective nullable fields.
+        no_unify: set[str] | None, default None
+            Prevent unification of keys under these field names with their sibling record fields.
         force_field_types : dict[str, str], optional
             Per-field overrides for schema inference (e.g. ``{"labels": "map"}``).
+        wrap_scalars : bool, default True
+            Whether to promote scalar values into singleton objects when they appear
+            in contexts where other rows provide objects. This avoids unification
+            failures between scalars and objects. The promoted field name defaults
+            to the parent key with a ``__{type}`` suffix, e.g. a string under
+            ``"value"`` becomes ``{"value__string": "..."}``.
         wrap_root : str | bool | None, default None
             If a string, wrap each JSON row under that key before normalisation.
             If ``True``, wrap under the column name. If ``None``, leave rows unchanged.
+        no_root_map : bool, default True
+            Prevent document root from becoming a map type, even if it meets map inference criteria
 
         Returns:
         -------
@@ -533,27 +739,43 @@ class GensonNamespace:
             empty_as_null=empty_as_null,
             coerce_strings=coerce_strings,
             map_encoding=map_encoding,
+            profile=profile,
             map_threshold=map_threshold,
+            map_max_required_keys=map_max_required_keys,
+            unify_maps=unify_maps,
             force_field_types=force_field_types,
+            wrap_scalars=wrap_scalars,
             wrap_root=wrap_root_field,
+            no_root_map=no_root_map,
         )
         if decode:
             if map_encoding != "kv":
                 # Map type fields must be k:v encoded as infer_polars_schema assumes it
                 # This could be done, it would always make record fields, ...but why?
                 raise NotImplementedError("map_encoding must be kv to decode to Polars")
-            # Infer Avro schema and convert it to Polars Schema
-            schema = self.infer_polars_schema(
-                column,
-                ignore_outer_array=ignore_outer_array,
-                ndjson=ndjson,
-                merge_schemas=True,
-                map_threshold=map_threshold,
-                force_field_types=force_field_types,
-                avro=True,
-                wrap_root=wrap_root_field,
-            )
-            dtype = pl.Struct(schema)
+
+            if decode is True:
+                # Infer Avro schema and convert it to Polars Schema
+                schema = self.infer_polars_schema(
+                    column,
+                    ignore_outer_array=ignore_outer_array,
+                    ndjson=ndjson,
+                    merge_schemas=True,
+                    profile=profile,
+                    map_threshold=map_threshold,
+                    map_max_required_keys=map_max_required_keys,
+                    unify_maps=unify_maps,
+                    force_field_types=force_field_types,
+                    wrap_scalars=wrap_scalars,
+                    avro=True,
+                    wrap_root=wrap_root_field,
+                    no_root_map=no_root_map,
+                )
+                dtype = pl.Struct(schema)
+            else:
+                # decode was passed as a Polars Schema directly
+                dtype = decode
+
             result = self._df.select(expr.str.json_decode(dtype=dtype))
             if unnest:
                 result = result.unnest(expr.meta.output_name())
